/***
 * Julia
 * Copyright (C) 2005-2006 INRIA, France Telecom, USTL
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Lionel.Seinturier@lifl.fr
 *
 * Author: Lionel Seinturier
 */

package org.objectweb.fractal.juliak.asm;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;

import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.julia.InitializationContext;
import org.objectweb.fractal.julia.InterceptorInterface;
import org.objectweb.fractal.julia.asm.CodeGenerator;
import org.objectweb.fractal.julia.loader.Tree;
import org.objectweb.fractal.juliak.platform.Platform;


/**
 * This class generates {@link org.objectweb.fractal.julia.Interceptor}
 * implementations.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@lifl.fr>
 * @since 2.5
 */
public class InterceptorClassGenerator extends AbstractClassGenerator {
    
    /**
     * The descriptor for this generator.
     * 
     * The string representation of this {@link Tree} structure follows a LISP
     * like syntax such as "(foo (bar foo))".
     * 
     * An example of such a descriptor for lifecycle interceptors follows:
     * "(org.objectweb.fractal.julia.asm.InterceptorClassGenerator
     *    org.objectweb.fractal.julia.asm.LifeCycleCodeGenerator
     * )"
     * 
     * Another, slightly more complex one, taken from the statController
     * example:
     * (org.objectweb.fractal.julia.asm.InterceptorClassGenerator
     *    org.objectweb.fractal.julia.asm.LifeCycleCodeGenerator
     *    stat.StatCodeGenerator
     *    stat.StatClassTransformer # transforms field accesses in content class
     * )
     */
    private Tree tdesc;
    
    public InterceptorClassGenerator( String desc ) throws TreeParserException {
        TreeParser tp = new TreeParser(desc);
        tdesc = tp.parseTree();
    }
    
    /**
     * Return the name of the suffix which is appended to generated class names.
     */
    protected String getClassNameSuffix( InterfaceType it ) {
        Tree[] trees = tdesc.getSubTrees();
        // Skip the 1st sub tree (InterceptorClassGenerator)
        String suffix = "";
        for (int i = 1; i < trees.length; i++) {
            String cgclassname = trees[i].toString();
            int lastdot = cgclassname.lastIndexOf('.');
            suffix +=
                (lastdot==-1) ? cgclassname : cgclassname.substring(lastdot+1);
        }
        suffix += it.isFcClientItf() ? "Out" : "In";
        return suffix;
    }

    /**
     * Return true if the mode (IN, OUT or IN_OUT) of at least one of the class
     * generators matchs the mode (client or server) of the given interface
     * type.
     */
    public boolean match( InterfaceType it ) {
        Tree[] trees = tdesc.getSubTrees();
        // Skip the 1st sub tree (InterceptorClassGenerator)
        for (int i = 1; i < trees.length; i++) {
            String cgclassname = trees[i].toString();
            try {
                Class cl = Platform.get().loadClass(cgclassname);
                CodeGenerator cg = (CodeGenerator) cl.newInstance();
                int mode = cg.init(null);
                if( mode==CodeGenerator.IN && ! it.isFcClientItf() ) {
                    return true;
                }
                if( mode==CodeGenerator.IN_OUT ) {
                    return true;
                }
                if( mode==CodeGenerator.OUT && it.isFcClientItf() ) {
                    return true;
                }
            }
            catch (ClassNotFoundException e) {
                final String msg = e.getClass()+": "+e.getMessage();
                throw new RuntimeException(msg);
            } catch (InstantiationException e) {
                final String msg = e.getClass()+": "+e.getMessage();
                throw new RuntimeException(msg);
            } catch (IllegalAccessException e) {
                final String msg = e.getClass()+": "+e.getMessage();
                throw new RuntimeException(msg);
            }
        }
        
        return false;
    }

    /**
     * Generate the bytecode of an interceptor class implementing the given
     * Fractal interface type.
     * 
     * @param targetClassname    the name of the class to generate
     * @param it  the corresponding interface type
     * @param ic  the context for initializing the generated class
     * @return    the bytecode of the generated class or
     *            null if the interceptor does not apply to the current
     *            interface type
     */
    public byte[] generate(
            String targetClassname, InterfaceType it,
            InitializationContext ic )
    throws org.objectweb.fractal.api.factory.InstantiationException {
                
        /*
         * Retrieve controller classes.
         * They are needed by the Julia bytecode generation framework.
         * The rationale is that the lifecycle interceptor invokes not only the
         * public interface of the lifecycle controller, but also modifies some
         * public fields (e.g. fcInvocationCounter). We can then not rely on
         * control interfaces, and we need implementation classes.
         */
        List ctrlclasses = new ArrayList();
        Collection values = ic.interfaces.values();
        for (Iterator iter = values.iterator(); iter.hasNext();) {
            Object  o = iter.next();
            Class cl = o.getClass();
            ctrlclasses.add(cl);
        }
        Tree[] ctrls = new Tree[ctrlclasses.size()];
        int i = 0;
        for (Iterator iter = ctrlclasses.iterator(); iter.hasNext();) {
            Class ctrlclass = (Class) iter.next();
            ctrls[i] = new Tree(ctrlclass.getName());
            i++;
        }
        
        /*
         * Use the Julia bytecode generation framework for interceptors.
         * 
         * t typically looks like:
         *     (org.objectweb.fractal.julia.asm.LifeCycleCodeGenerator) or
         *     (org.objectweb.fractal.julia.asm.LifeCycleCodeGenerator
         *          stat.StatCodeGenerator
         *          stat.StatClassTransformer
         *     )
         */
        org.objectweb.fractal.julia.asm.InterceptorClassGenerator icg =
            new org.objectweb.fractal.julia.asm.InterceptorClassGenerator();
        Tree[] t = new Tree[tdesc.getSize()-1];
        System.arraycopy(tdesc.getSubTrees(), 1, t, 0, t.length);
        icg.initialize(new Tree(t));
        
        /*
         * Build the {@link Tree} structure needed as input by the Julia
         * bytecode generation framework.
         */
        Class superclass =
            it.isFcClientItf() ? InterceptorInterface.class : Object.class;
        String signature = it.getFcItfSignature();
        Tree args =
            new Tree(
                new Tree[]{
                    tdesc,
                    new Tree(superclass.getName()),
                    new Tree( new Tree[]{new Tree(signature)} ),
                    new Tree(ctrls),
                    new Tree(it.isFcClientItf() ? "out" : "in")
                }
            );
        
        byte[] bytecode = null;
        try {
            bytecode =
                icg.generateClass( targetClassname, args, getLoader(), getClassLoader() );
        }
        catch( Exception cge ) {
            /*
             * Actual thrown exception is
             * org.objectweb.julia.asm.ClassGenerationException. Catch Exception
             * to avoid introducing a dependency towards julia-asm.jar. It seems
             * that caught exceptions introduce a dependency as soon as the
             * class even if the code is not reached.
             */
            final String msg = "ClassGenerationException: "+cge.getMessage();
            throw new org.objectweb.fractal.api.factory.InstantiationException(msg);
        }

        return bytecode;
    }
}
