/***
 * Julia
 * Copyright (C) 2005-2006 INRIA, France Telecom, USTL
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Lionel.Seinturier@lifl.fr
 *
 * Author: Lionel Seinturier
 */

package org.objectweb.fractal.juliak.control.interceptor;

import java.util.HashMap;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.ContentController;
import org.objectweb.fractal.api.factory.InstantiationException;
import org.objectweb.fractal.api.type.ComponentType;
import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.julia.ComponentInterface;
import org.objectweb.fractal.julia.Controller;
import org.objectweb.fractal.julia.InitializationContext;
import org.objectweb.fractal.julia.Interceptor;
import org.objectweb.fractal.juliak.FractalHelper;
import org.objectweb.fractal.juliak.asm.ClassGenerator;
import org.objectweb.fractal.juliak.asm.InterceptorClassGenerator;
import org.objectweb.fractal.juliak.asm.TreeParserException;
import org.objectweb.fractal.juliak.control.binding.BindingControllerDef;
import org.objectweb.fractal.juliak.control.content.ContentControllerDef;
import org.objectweb.fractal.juliak.factory.InstanceGenerator;

/**
 * This class implements an interceptor controller for managing the
 * interceptors associated to the client and server interfaces of a component.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@lifl.fr>
 * @since 2.5
 */
public abstract class InterceptorControllerMixin
    implements Controller, InterceptorController, InterceptorDefAttributes {

    // -----------------------------------------------------------------
    // Implementation of the Controller interface
    // -----------------------------------------------------------------
    
    public void initFcController(InitializationContext ic)
    throws InstantiationException {
        
        if( interceptorClassGenerator != null ) {
            
            /*
             * Prepare a context for initializing interceptors.
             * Julia generated interceptors assume ic.interfaces directly
             * references the controller implementations.
             */
            BindingController bc = (BindingController)
                ic.interfaces.get("//"+BindingControllerDef.NAME);
            String[] itfNames = bc.listFc();
            InitializationContext intic = new InitializationContext();
            intic.interfaces = new HashMap();
            for (int i = 0; i < itfNames.length; i++) {
                String itfName = itfNames[i].substring(2);  // Strip //
                Object end = FractalHelper.lookupFc(bc, itfNames[i]);
                while( end instanceof ComponentInterface ) {
                    ComponentInterface ci = (ComponentInterface) end;
                    end = ci.getFcItfImpl();
                }
                intic.interfaces.put(itfName,end);
            }
            
            Component compctrl = (Component) ic.interfaces.get("component");
            ContentController cc = null;
            try {
                Object o = compctrl.getFcInterface(ContentControllerDef.NAME);
                cc = (ContentController) o;
            }
            catch( NoSuchInterfaceException nsie ) {}
            
            /*
             * Retrieve the list of fcinterfaces associated to the base level
             * component and generate the needed interceptors.
             */
            ComponentType ct = (ComponentType) compctrl.getFcType();
            InterfaceType[] its = ct.getFcInterfaceTypes();
            for (int i = 0; i < its.length; i++) {
                InterfaceType it = its[i];
                String itfName = it.getFcItfName();
                if( it.isFcCollectionItf() ) {
                    itfName = "/juliak/collection/" + it.getFcItfName();
                }
                Object itf = FractalHelper.getFcInterface(compctrl,itfName);
                addFcInterceptor(itf, intic, cc);
            }
        }

        _super_initFcController(ic);
    }
    
    private void addFcInterceptor(
            Object o, InitializationContext intic, ContentController cc )
    throws InstantiationException {
        
        ComponentInterface itf = (ComponentInterface) o;
        InterfaceType it = (InterfaceType) itf.getFcItfType();
        String itfName = itf.getFcItfName();
        
        if( FractalHelper.isFcControllerInterfaceName(itfName) ) {
            return;
        }
        
        /*
         * Check that the mode (IN, OUT or IN_OUT) of the class
         * generator matchs the mode (client or server) of the interface
         * type.
         */
        if( ! interceptorClassGenerator.match(it) ) {
            return;
        }
        
        /*
         * Create the interceptor.
         * 
         * generate returns null if the generation is not appropriate
         * e.g. the code generator mode (IN, OUT, or IN_OUT) does not
         * apply for this interface type.
         */
        Object content = interceptorInstanceGenerator.generate(it,intic);
        Interceptor interceptor = (Interceptor) content;
        interceptor.initFcController(intic);
        
        // Add the interceptor
        addFcInterceptor(itf, interceptor);
        
        // The same for internal interfaces
        if( cc != null ) {
            if( it.isFcCollectionItf() ) {
                itfName = "/juliak/collection/" + itfName;
            }
            ComponentInterface intitf = (ComponentInterface)
                FractalHelper.getFcInternalInterface(cc, itfName);
            addFcInterceptor(intitf, interceptor);
        }
    }
    
    // -----------------------------------------------------------------
    // Implementation of the InterceptorController interface
    // -----------------------------------------------------------------
    
    /**
     * Appends the specified interceptor at the end of the list of already
     * existing interceptors for the specified Fractal interface.
     * 
     * @param itf          the Fractal interface
     * @param interceptor  the interceptor
     */
    public void addFcInterceptor( ComponentInterface itf, Interceptor interceptor ) {
        
        /*
         * Fetch the last interceptor interceptor in the list of interceptors
         * already associated to the specified interface.
         */
        Object last = null;
        Object next = itf.getFcItfImpl();
        while( next instanceof Interceptor ) {
            Interceptor i = (Interceptor) next;
            last = next;
            next = i.getFcItfDelegate();
        }
        
        if( last == null ) {
            // The list of existing interceptors is empty
            Object content = itf.getFcItfImpl();
            interceptor.setFcItfDelegate(content);
            itf.setFcItfImpl(interceptor);
        }
        else {
            Interceptor i = (Interceptor) last;
            Object content = i.getFcItfDelegate();
            interceptor.setFcItfDelegate(content);
            i.setFcItfDelegate(interceptor);
        }
        
        itf.updateFcState();                                
    }

    /**
     * Returns the array of interceptors associated to the specified Fractal
     * interface.
     * 
     * @param itf  the Fractal interface
     * @return     the array of interceptors
     */
    public Interceptor[] getFcInterceptors( ComponentInterface itf ) {

        Object next = itf.getFcItfImpl();
        int size = 0;
        while( next instanceof Interceptor ) {
            Interceptor i = (Interceptor) next;
            next = i.getFcItfDelegate();
            size++;
        }
        
        Interceptor[] result = new Interceptor[size];
        next = itf.getFcItfImpl();
        size = 0;
        while( next instanceof Interceptor ) {
            Interceptor i = (Interceptor) next;
            result[size] = i;
            next = i.getFcItfDelegate();
            size++;
        }
        
        return result;
    }
    
    /**
     * Removes the specified interceptor from the list of already existing
     * interceptors for the specified Fractal interface.
     * 
     * @param itf          the Fractal interface
     * @param interceptor  the interceptor
     * @return             true if the list contained the specified interceptor
     */
    public boolean removeFcInterceptor( ComponentInterface itf, Interceptor interceptor ) {

        Object previous = null;
        Object next = itf.getFcItfImpl();
        while( next instanceof Interceptor ) {
            Interceptor i = (Interceptor) next;
            if( next == interceptor ) {
                // The interceptor has been found
                Object nextnext = i.getFcItfDelegate();
                if( previous == null ) {
                    // The interceptor is the first in the list
                    itf.setFcItfImpl(nextnext);
                }
                else {
                    Interceptor p = (Interceptor) previous;
                    p.setFcItfDelegate(nextnext);
                }
                itf.updateFcState();                                
                return true;
            }
            
            previous = next;
            next = i.getFcItfDelegate();
        }
        
        // The interceptor hasn't been found
        return false;
    }
    
    
    // -----------------------------------------------------------------
    // Implementation of the InterceptorDefAttributes interface
    // -----------------------------------------------------------------
    
    /**
     * @param interceptors
     *      A stringified representation of a Julia
     *      {@link org.objectweb.fractal.julia.loader.Tree} structure describing
     *      the interceptor class generators associated with this control
     *      membrane.
     */
    public void setInterceptors( String interceptors ) {    
        this.interceptors = interceptors;
        
        try {
            interceptorClassGenerator =
                new InterceptorClassGenerator(interceptors);
        }
        catch (TreeParserException e) {
            final String msg =
                "Bad interceptor descriptor: "+e.getMessage()+"\n"+interceptors;
            throw new RuntimeException(msg);
        }
        
        interceptorInstanceGenerator =
            new InstanceGenerator(interceptorClassGenerator);
    }
    
    /**
     * @return
     *      A stringified representation of a Julia
     *      {@link org.objectweb.fractal.julia.loader.Tree} structure describing
     *      the interceptor class generators associated with this control
     *      membrane.
     */
    public String getInterceptors() {
        return interceptors;
    }

    private String interceptors = "";    
    private ClassGenerator interceptorClassGenerator;
    private InstanceGenerator interceptorInstanceGenerator;
    
    
    // -------------------------------------------------------------------------
    // Fields and methods required by the mixin class in the base class
    // -------------------------------------------------------------------------

    abstract public void _super_initFcController( InitializationContext ic )
    throws InstantiationException;
}
