/***
 * Julia
 * Copyright (C) 2005-2006 INRIA, France Telecom, USTL
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Lionel.Seinturier@lifl.fr
 *
 * Author: Lionel Seinturier
 */

package org.objectweb.fractal.juliak.factory;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.Interface;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.Type;
import org.objectweb.fractal.api.control.ContentController;
import org.objectweb.fractal.api.factory.GenericFactory;
import org.objectweb.fractal.api.factory.InstantiationException;
import org.objectweb.fractal.api.type.ComponentType;
import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.api.type.TypeFactory;
import org.objectweb.fractal.julia.Controller;
import org.objectweb.fractal.julia.InitializationContext;
import org.objectweb.fractal.julia.asm.ClassGenerator;
import org.objectweb.fractal.julia.asm.MixinClassGenerator;
import org.objectweb.fractal.julia.loader.Loader;
import org.objectweb.fractal.julia.loader.Tree;
import org.objectweb.fractal.julia.type.BasicComponentType;
import org.objectweb.fractal.julia.type.BasicInterfaceType;
import org.objectweb.fractal.juliak.FractalHelper;
import org.objectweb.fractal.juliak.Juliak;
import org.objectweb.fractal.juliak.membrane.MCompositeImpl;
import org.objectweb.fractal.juliak.membrane.MPrimitiveImpl;
import org.objectweb.fractal.juliak.membrane.MembraneDef;
import org.objectweb.fractal.juliak.membrane.Membranes;


/**
 * A generic factory which instanciates components with a component-based
 * membrane.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@lifl.fr>
 * @since 2.5
 */
public class BasicGenericFactoryImpl
    extends AbstractJuliaEnabledGenericFactoryImpl
    implements GenericFactory {
    
    // -----------------------------------------------------------------
    // Membrane related methods
    // -----------------------------------------------------------------
        
    /**
     * Create a new control membrane.
     * 
     * @param type  the type of the component associated to this membrane
     * @param controllerDesc
     *          the controller description of this membrane
     *          (typically a String such as "primitive" or any of the other
     *          registered membrane identifier)
     * @param contentDesc 
     *          the content description of the component associated
     *          to this membrane (e.g. the fully qualified name of the class
     *          implementing a primitive component)
     * @param content
     *          the instance implementing the component associated
     *          to this membrane
     *          
     * @return  the fcinterface for the component interface
     *          exported by the given control membrane  
     */
    protected Component newFcMembrane(
            Type type, Object controllerDesc, Object contentDesc,
            Object content )
    throws InstantiationException {
        
        /*
         * Control membrane for controllers, i.e. components associated to the
         * mPrimitive and mComposite controller descriptions.
         */
        if( controllerDesc.equals("mPrimitive") ) {
            return new MPrimitiveImpl(type,content);
        }
        if( controllerDesc.equals("mComposite") ) {
            return new MCompositeImpl(type,content);
        }
        
        /*
         * Retrieve the membrane definition.
         */
        Membranes mr = Membranes.get();
        
        if( ! mr.containsControllerDesc(controllerDesc) ) {
            String msg = "Unknown controller description: "+controllerDesc;
            throw new InstantiationException(msg);
        }
        
        MembraneDef mdef = mr.getMembraneDef(controllerDesc);
        TypeFactory tf = new Juliak().getFcTypeFactory();
        GenericFactory gf = this;
        Component membrane = mdef.newFcStaticMembrane(tf,gf);                
        
        /*
         * Bind the membrane to the content.
         */
        Component compctrlitf =
            bindFcMembrane(type,contentDesc,content,membrane);

        return compctrlitf;
    }
    
    /**
     * Bind a control membrane to the component defined by the given type,
     * content description and content.
     * 
     * @param type  the type of the component associated to this membrane
     * @param contentDesc 
     *          the content description of the component associated
     *          to this membrane (e.g. the fully qualified name of the class
     *          implementing a primitive component)
     * @param content
     *          the instance implementing the component associated
     *          to this membrane
     * @param membrane  the control membrane
     * 
     * @return  the fcinterface for the component interface
     *          exported by the given control membrane  
     */
    protected Component bindFcMembrane(
            Type type, Object contentDesc, Object content, Component membrane )
    throws InstantiationException {
        
        /*
         * Check that the membrane fits all the requirements.
         */
        checkFcMembrane(membrane);
        
        /*
         * Create the initialisation context.
         */
        InitializationContext ic = new InitializationContext();
        ic.controllers = new ArrayList();
        ic.interfaces = new HashMap();
        ic.internalInterfaces = new HashMap();
        
        ic.interfaces.put("/content",content);
        
        /*
         * In the case of template components, Julia factory controller expects
         * ic.content to reference contentDesc (which is an array describing
         * the component associated to the template). contentDesc may be null
         * for composite components.
         */
        if( contentDesc == null ) {
            ic.content = content;
        }
        else {
            ic.content =
                (contentDesc.getClass().isArray()) ? contentDesc : content;
        }

        /*
         * Retrieve all controllers.
         */
        Map ctrls = new HashMap();
        List l = new ArrayList();
        FractalHelper.addAllSubComponents(membrane,l);
        for (Iterator iter = l.iterator(); iter.hasNext();) {
            Component comp = (Component) iter.next();
            
            // composite control components may have a null content
            Object ctrl = FractalHelper.getContent(comp);
            if( ctrl != null ) {
                ic.controllers.add(ctrl);
                ctrls.put(ctrl,comp);
            }
        }
        
        /*
         * Initialize the interface and the interface type of the component
         * controller.
         */
        Interface compctrlitf = (Interface)
            FractalHelper.getFcInterface(membrane,"//component");
        Component compctrl = (Component) compctrlitf;
        InterfaceType srcit = (InterfaceType) compctrlitf.getFcItfType();
        InterfaceType it = downToLevel0InterfaceType(srcit);
        InterfaceInstanceGenerator itfgen = InterfaceInstanceGenerator.get();
        Interface proxy = itfgen.generate(it,compctrl,compctrl,false);
        Component proxyForCompctrl = (Component) proxy;
        ic.interfaces.put( "component", proxyForCompctrl );
        
        List fullITs = new ArrayList();
        fullITs.add(it);
        
        /*
         * Proceed by initializing the interface and the interface type of the
         * other controllers.
         */
        Object[] itfs = membrane.getFcInterfaces();
        for (int i = 0; i < itfs.length; i++) {
            Interface itf = (Interface) itfs[i];
            String name = itf.getFcItfName();
            
            if( name.startsWith("//") && !name.equals("//component") ) {
                
                srcit = (InterfaceType) itf.getFcItfType();
                it = downToLevel0InterfaceType(srcit);
                
                // Control (external) interface
                proxy = itfgen.generate(it,proxyForCompctrl,itf,false);
                name = name.substring(2);    // strip //
                ic.interfaces.put( name, proxy );

                // Control (internal) interface
                if( name.equals("factory") ) {
                    InterfaceType intit = newSymetricInterfaceType(it);
                    Interface internal = itfgen.generate(intit,proxyForCompctrl,itf,true);
                    ic.internalInterfaces.put(name,internal);
                }

                // Component type
                if( name.charAt(0) != '/' ) {
                    fullITs.add(it);
                }
            }
        }
        
        /*
         * Build the component type by appending the business type to the
         * already computed control type.
         */
        ComponentType comptype = (ComponentType) type;
        InterfaceType[] its = comptype.getFcInterfaceTypes();
        fullITs.addAll( Arrays.asList(its) );
        
        InterfaceType[] full = new InterfaceType[fullITs.size()];
        ic.type = new BasicComponentType( (InterfaceType[]) fullITs.toArray(full) );

        /*
         * Build business component interfaces and interceptors.
         */
        for (int i = 0; i < its.length; i++) {
            
            String name = its[i].getFcItfName();
            
            if( its[i].isFcCollectionItf() ) {
                // Julia naming convention for collection interfaces
                name = "/collection/"+name;
            }
            
            // External interface
            Interface ext = itfgen.generate(its[i],proxyForCompctrl,content,false);
            ic.interfaces.put(name,ext);
            if( its[i].isFcCollectionItf() ) {
                /*
                 * Add a 2nd reference for this interface.
                 * It will be possible to retrieve this reference to dynamically
                 * manage (add and remove) the interceptors associated to a
                 * collection interface. This is not possible with
                 * "/collection/"+name which is never returned as this by
                 * getFcInterface (a clone is always returned).
                 */
                ic.interfaces.put("/juliak"+name,ext);
            }
            
            // Internal interface
            if( ! name.equals("attribute-controller") ) {
                InterfaceType intit = newSymetricInterfaceType(its[i]);
                Interface internal =
                    itfgen.generate(intit,proxyForCompctrl,content,true);
                ic.internalInterfaces.put(name,internal);
                if( its[i].isFcCollectionItf() ) {
                    ic.internalInterfaces.put("/juliak"+name,internal);
                }
            }
        }

        /*
         * Initialize controllers.
         * 
         * Iterate on ic.controllers which is a List, to initialize the
         * component controller first, and the ContentController second.
         * When initialized, controllers (e.g. the interceptor controller) may
         * assume that the component controller has already been initialized to
         * retrieve for example, the component type or the array of internal
         * fcinterfaces.
         */
        // First the Component control-component
        for (Iterator iter = ic.controllers.iterator(); iter.hasNext();) {
            Object o = iter.next();
            if( o instanceof Component ) {
                Controller ctrl = (Controller) o;
                ctrl.initFcController(ic);
            }
        }
        
        // Second the Content control-component
        for (Iterator iter = ic.controllers.iterator(); iter.hasNext();) {
            Object o = iter.next();
            if( o instanceof ContentController ) {
                Controller ctrl = (Controller) o;
                Component ctrlcomp = (Component) ctrls.get(ctrl);
                InitializationContext localic = new InitializationContext();
                localic.content = ic.content;
                localic.type = ic.type;
                localic.interfaces = new ControlComponentBoundMap(ctrlcomp);
                localic.internalInterfaces = ic.internalInterfaces;
                ctrl.initFcController(ic);
            }
        }
        
        // Remaining control-components
        for (Iterator iter = ic.controllers.iterator(); iter.hasNext();) {
            Object o = iter.next();

            if( o instanceof Component || o instanceof ContentController ) {
                continue;
            }

            if( !(o instanceof Controller) ) {
                /*
                 * This is not mandatory for controllers to implement
                 * {@link Controller}. For example, simple controllers which do
                 * not require any information from the initialization context
                 * may avoid implementing {@link Controller}.
                 */
                continue;
            }
            
            Controller ctrl = (Controller) o;
            Component ctrlcomp = (Component) ctrls.get(ctrl);

            InitializationContext localic = new InitializationContext();
            localic.content = ic.content;
            localic.type = ic.type;
            
            /*
             * It seems that Julia controllers do not use ic.controllers.
             * This list seems to be only used by the generic factory.
             */
            localic.controllers = null;
            localic.interfaces = new ControlComponentBoundMap(ctrlcomp);
            localic.internalInterfaces = null;                    
            
            ctrl.initFcController(localic);
        }
        
        return proxyForCompctrl;
    }

    
    // -----------------------------------------------------------------
    // Implementation specific
    // -----------------------------------------------------------------
        
    /**
     * Load the specified content class.
     * First, try to load the class from the classpath.
     * If that fails, assume that the specified contentDesc corresponds to an
     * alias defined in the julia.cfg configuration file. This is case for
     * control component where the content class is dynamically generated by
     * the ASM mixin class generator.
     * 
     * @param controllerDesc
     *      the controller descriptor of the component for which the content
     *      class is requested
     * @param contentDescStr
     *      the content descriptor. May be a fully-qualified class name or the
     *      stringified representation of a {@link Tree} to be loaded by a Julia
     *      {@link Loader}. 
     */
    protected Class loadFcContentClass( Object controllerDesc, String contentDescStr )
    throws InstantiationException {

        try {
            Class contentClass = super.loadFcContentClass(controllerDesc, contentDescStr);
            return contentClass;
        }
        catch(InstantiationException ie) {
            
            /*
             * This is not the fully qualified name of a class.
             * Try to load the corresponding Tree from julia.cfg.
             */
            
            Loader loader = Juliak.get().getFcLoader();
            ClasseDefinitions clrep = ClasseDefinitions.get();
            try {
                Tree src = loader.loadTree(contentDescStr);
                src = src.getSubTree(0);
                String targetClassname = src.getSubTree(1).toString();
                
                /*
                 * Check whether the class has already been requested.
                 */
                if( clrep.containsClass(targetClassname) ) {
                    ClassDefinition cd = clrep.getClass(targetClassname);
                    return cd.cl;
                }
                
                /*
                 * Mixin the layers and register the generated class.
                 */
                byte[] bytecode =
                    mixincg.generateClass(targetClassname, src, loader, cl);
                ClassDefinition cd =
                    clrep.registerClass(targetClassname, bytecode);
                
                return cd.cl;
            }
            catch (Exception e) {
                final String msg = "Content class "+contentDescStr+" not found";
                throw new InstantiationException(msg);
            }
        }
    }
    
    private ClassGenerator mixincg = new MixinClassGenerator();
    private ClassLoader cl = getClass().getClassLoader();

    /**
     * Check that the given control membrane fits all the requirements.
     */
    private void checkFcMembrane( Component membrane ) throws InstantiationException {
        
        try {
            membrane.getFcInterface("//component");
        }
        catch (NoSuchInterfaceException nsie) {
            final String msg =
                "The given membrane should provide a "+nsie.getMessage()+
                " interface";
            throw new InstantiationException(msg);
        }
    }
    
    /**
     * Given the interface type of a control membrane, return the corresponding
     * type for the application (level 0) interface. This consists in removing
     * the 2 slash characters which are in front of the interface name.
     */
    private static InterfaceType downToLevel0InterfaceType( InterfaceType srcit ) {
        InterfaceType it =
            new BasicInterfaceType(
                    srcit.getFcItfName().substring(2),     // remove //
                    srcit.getFcItfSignature(),
                    srcit.isFcClientItf(),
                    srcit.isFcOptionalItf(),
                    srcit.isFcCollectionItf()
            );
        return it;
    }
    
    /**
     * Return a copy of the given interface type where the client role is
     * transformed into server and the server role is transformed into client.
     */
    private static InterfaceType newSymetricInterfaceType( InterfaceType it ) {
        return
            new BasicInterfaceType(
                    it.getFcItfName(),
                    it.getFcItfSignature(),
                    ! it.isFcClientItf(),   // client <-> server
                    it.isFcOptionalItf(),
                    it.isFcCollectionItf()
            );        
    }
    
}
